#include "SysParam.h"
#include "accessControl/userControl.h"
#include "ml/mlModule.h"

#include <iostream>

using std::cout;
using std::endl;

extern void runUnitTest();

void printHelp() {

    string helpMsg = "This is the world's most secure IoT machine learning system, the syntax to use the tools are as follows:\n"
                     "./embeddedlab [operation category][operation][parameter 1]...[parameter n][username][password][user database path]\n\n"
                     "account operations:\n"
                     "initialize new user database: ./embeddedlab account init [path to user database]\n"
                     "authenticate user: ./embeddedlab account authenticate [username][password][path to user database]\n"
                     "add a new user: ./embeddedlab account add [username to add][password of the new user][username][password][path to user database]\n"
                     "remove a existing user: ./embeddedlab account remove [username to remove][username][password][path to user database]\n\n"
                     "ml operations:\n"
                     "train a model: ./embeddedlab ml train [path to data][path to save model][username][password][path to user database]\n"
                     "inference: ./embeddedlab ml inference [path to test data][path to saved model][username][password][path to user database]\n\n"
                     "PLEASE NOTE: IN REAL LIFE, IT IS A BAD IDEA TO PASS IN PASSWORD AS A COMMAND LINE ARGUMENT\n";
    cout << helpMsg << endl;
}


int main(int argc, char *argv[]) {

    // run unit test if no argument is given
    if (argc == UNIT_TEST) {
        runUnitTest();
        return SUCCESS;
    }

    if (argc < MIN_NUMBER_OF_ARGUMENTS || argc > MAX_NUMBER_OF_ARGUMENTS) {
        printHelp();
        return WRONG_ARGUMENTS;
    }

    if (strncmp(argv[LIBRARY], "help", sizeof("help")) == 0) {
        printHelp();
        return SUCCESS;
    }

    if (strncmp(argv[LIBRARY], "account", sizeof("account")) == 0) {
        cout << "account operation" << endl;

        // user database initialization
        if (strncmp(argv[FUNCTION], "init", sizeof("init")) == 0) {

            // check if db path is given
            if (argc < MIN_NUMBER_OF_ARGUMENTS + 1) {
                printHelp();
                return WRONG_ARGUMENTS;
            }

            // get db path
            const char *dbPath = argv[PARAM1];

            // check if db is already exist
            ifstream db(dbPath);
            if (db.good()) {
                cout << "user database already exist, please log in." << endl;
                return SUCCESS;
            }

            // initialize new user database
            UserDatabase::getInstance().loadUserDatabase(dbPath);
            cout << "user database initialized." << endl;
            cout << "a default user 'admin' with password 'admin' is created." << endl;
        }

        // authentication
        if (strncmp(argv[FUNCTION], "authenticate", sizeof("authenticate")) == 0) {

            // check if username and password are given
            if (argc != MIN_NUMBER_OF_ARGUMENTS + 3) {
                printHelp();
                return WRONG_ARGUMENTS;
            }

            const char *username = argv[PARAM1];
            const char *password = argv[PARAM2];
            const char *dbPath = argv[PARAM3];

            UserDatabase::getInstance().loadUserDatabase(dbPath);

            if (UserDatabase::getInstance().authenticateUser(username, password) == 0) {
                cout << "user '" << username << "' is authorized." << endl;
                return SUCCESS;
            }
            cout << "authentication failed, please check your username and password." << endl;
            return AUTH_FAILED;
        }

        // add new user
        if (strncmp(argv[FUNCTION], "add", sizeof("add")) == 0) {

            // check if username and password are given
            if (argc != MIN_NUMBER_OF_ARGUMENTS + 5) {
                printHelp();
                return WRONG_ARGUMENTS;
            }

            const char *usernameToAdd = argv[PARAM1];
            const char *passwordToAdd = argv[PARAM2];

            const char *username = argv[PARAM3];
            const char *password = argv[PARAM4];
            const char *dbPath = argv[PARAM5];

            UserDatabase::getInstance().loadUserDatabase(dbPath);
            UserDatabase::getInstance().authenticateUser(username, password);

            if (!UserDatabase::getInstance().isAuthenticated()) {
                cout << "user is not authorized, please log in first." << endl;
                return AUTH_FAILED;
            }

            if (UserDatabase::getInstance().addUser(usernameToAdd, passwordToAdd) != 0) {
                cout << "user " << usernameToAdd << " already exists." << endl;
                cout << "please try another user name." << endl;
                return ADD_USER_FAILED;
            }

            cout << "user '" << usernameToAdd << "' is added to the database." << endl;
        }

        // remove a exist user
        if (strncmp(argv[FUNCTION], "remove", sizeof("remove")) == 0) {

            // check if username and password are given
            if (argc != MIN_NUMBER_OF_ARGUMENTS + 4) {
                printHelp();
                return WRONG_ARGUMENTS;
            }

            const char *usernameToRemove = argv[PARAM1];

            const char *username = argv[PARAM2];
            const char *password = argv[PARAM3];
            const char *dbPath = argv[PARAM4];

            UserDatabase::getInstance().loadUserDatabase(dbPath);

            UserDatabase::getInstance().authenticateUser(username, password);

            if (!UserDatabase::getInstance().isAuthenticated()) {
                cout << "user is not authorized, please log in first." << endl;
                return AUTH_FAILED;
            }

            if (UserDatabase::getInstance().removeUser(usernameToRemove) != 0) {
                cout << "user '" << username << "' does not exist in the database." << endl;
                cout << "please try another user name." << endl;
                return REMOVE_USER_FAILED;
            }

            cout << "user '" << usernameToRemove << "' is removed from the database." << endl;
        }

    }
    else if (strncmp(argv[LIBRARY], "ml", sizeof("ml")) == 0) {
        cout << "ml operation" << endl;

        // train and save model
        if (strncmp(argv[FUNCTION], "train", sizeof("train")) == 0) {
            // check if username and password are given
            if (argc != MIN_NUMBER_OF_ARGUMENTS + 5) {
                printHelp();
                return WRONG_ARGUMENTS;
            }

            const char *dataPath = argv[PARAM1];
            const char *modelPath = argv[PARAM2];

            const char *username = argv[PARAM3];
            const char *password = argv[PARAM4];
            const char *dbPath = argv[PARAM5];

            UserDatabase::getInstance().loadUserDatabase(dbPath);

            UserDatabase::getInstance().authenticateUser(username, password);

            if (!UserDatabase::getInstance().isAuthenticated()) {
                cout << "user is not authorized, please log in first." << endl;
                return AUTH_FAILED;
            }

            MlModule::getInstance()->trainMlModel(dataPath);

            MlModule::getInstance()->saveMlModel(modelPath);
        }

        // load model and inference
        if (strncmp(argv[FUNCTION], "inference", sizeof("inference")) == 0) {
            // check if username and password are given
            if (argc != MIN_NUMBER_OF_ARGUMENTS + 5) {
                printHelp();
                return WRONG_ARGUMENTS;
            }

            const char *dataPath = argv[PARAM1];
            const char *modelPath = argv[PARAM2];

            const char *username = argv[PARAM3];
            const char *password = argv[PARAM4];
            const char *dbPath = argv[PARAM5];

            UserDatabase::getInstance().loadUserDatabase(dbPath);

            UserDatabase::getInstance().authenticateUser(username, password);

            if (!UserDatabase::getInstance().isAuthenticated()) {
                cout << "user is not authorized, please log in first." << endl;
                return AUTH_FAILED;
            }

            MlModule::getInstance()->loadSavedMlModel(modelPath);

            MlModule::getInstance()->makeInference(dataPath);
        }
    } else {
        printHelp();
    }
    return 0;
}
