//
// Created by fangchenli on 7/22/19.
//


#include "ml/kmeans/kmeans.h"

#include <utility>
#include "SysParam.h"

using namespace std;

// reference:
// https://stackoverflow.com/questions/28287138/c-randomly-sample-k-numbers-from-range-0n-1-n-k-without-replacement/28287865#28287865

unordered_set<int> pickSet(int k, int n, mt19937 &gen) {
    unordered_set<int> elements;
    for (int r = n - k; r < n; ++r) {
        int v = uniform_int_distribution<>(1, r)(gen);
        if (!elements.insert(v).second) {
            elements.insert(r);
        }
    }
    return elements;
}

vector<int> pick(int k, int n) {
    random_device rd;
    mt19937 gen(rd());
    unordered_set<int> elements = pickSet(k, n, gen);
    vector<int> result(elements.begin(), elements.end());
    shuffle(result.begin(), result.end(), gen);
    return result;
}


Point::Point(int id, const string &line) {
    dimension = 0;
    pointId = id;
    stringstream is(line);
    double val;
    while (is >> val) {
        values.push_back(val);
        dimension++;
    }
    clusterId = 0;
}

int Point::getId() {
    return pointId;
}

int Point::getClusterId() {
    return clusterId;
}

int Point::getDimension() {
    return dimension;
}

double Point::getVal(int pos) {
    return values[pos];
}

void Point::setCluster(int val) {
    clusterId = val;
}


Cluster::Cluster(int clusterId, Point centroid) {
    this->clusterId = clusterId;
    for (int i = 0; i < centroid.getDimension(); ++i) {
        this->centroid.push_back(centroid.getVal(i));
    }
    this->addPoint(centroid);
}

void Cluster::addPoint(Point p) {
    p.setCluster(this->clusterId);
    this->points.push_back(p);
}

bool Cluster::removePoint(int pointId) {
    int size = points.size();

    for (int i = 0; i < size; i++) {
        if (points[i].getId() == pointId) {
            points.erase(points.begin() + i);
            return true;
        }
    }
    return false;
}

int Cluster::getId() {
    return clusterId;
}

Point Cluster::getPoint(int pos) {
    return points[pos];
}

int Cluster::getSize() {
    return points.size();
}

double Cluster::getCentroidByPos(int pos) {
    return centroid[pos];
}

void Cluster::setCentroidByPos(int pos, double val) {
    this->centroid[pos] = val;
}

void KMeans::setK(int k) {
    this->K = k;
}

void KMeans::setIters(int it) {
    this->iters = it;
}

void KMeans::setClusters(vector<Cluster> vc) {
    clusters = move(vc);
}

int KMeans::getNearestClusterId(Point p) {
    double sum = 0.0, min_dist;
    int NearestClusterId;

    for (int i = 0; i < p.dimension; i++) {
        sum += pow(clusters[0].getCentroidByPos(i) - p.getVal(i), 2.0);
    }

    min_dist = sqrt(sum);
    NearestClusterId = clusters[0].getId();

    for (int i = 1; i < K; i++) {
        double dist;
        sum = 0.0;

        for (int j = 0; j < p.dimension; j++) {
            sum += pow(clusters[i].getCentroidByPos(j) - p.getVal(j), 2.0);
        }

        dist = sqrt(sum);

        if (dist < min_dist) {
            min_dist = dist;
            NearestClusterId = clusters[i].getId();
        }
    }

    return NearestClusterId;
}

KMeans::KMeans() {
    this->K = DEFAULT_K;
    this->iters = DEFAULT_ITER;
}

KMeans::KMeans(int K, int iterations) {
    this->K = K;
    this->iters = iterations;
}

void KMeans::run(vector<Point> &points) {

    num_points = points.size();
    dimension = points[0].getDimension();

    //Initializing Clusters
    vector<int> initClusterIds = pick(K, num_points);
    vector<int> used_pointIds(initClusterIds.begin(), initClusterIds.end());
    for (int i = 0; i < K; ++i) {
        int index = initClusterIds[i];
        points[index].setCluster(i + 1);
        clusters.emplace_back(Cluster(i + 1, points[index]));
    }

//    cout << "Clusters initialized = " << clusters.size() << endl << endl;

//    cout << "Running K-Means Clustering.." << endl;

    int iter = 1;
    while (true) {
//        cout << "Iter - " << iter << "/" << iters << endl;
        bool done = true;

        // Add all points to their nearest cluster
        for (int i = 0; i < num_points; i++) {
            int currentClusterId = points[i].getClusterId();
            int nearestClusterId = getNearestClusterId(points[i]);

            if (currentClusterId != nearestClusterId) {
                if (currentClusterId != 0) {
                    for (int j = 0; j < K; j++) {
                        if (clusters[j].getId() == currentClusterId) {
                            clusters[j].removePoint(points[i].getId());
                        }
                    }
                }

                for (int j = 0; j < K; j++) {
                    if (clusters[j].getId() == nearestClusterId) {
                        clusters[j].addPoint(points[i]);
                    }
                }
                points[i].setCluster(nearestClusterId);
                done = false;
            }
        }

        // Recalculating the center of each cluster
        for (int i = 0; i < K; i++) {
            int ClusterSize = clusters[i].getSize();

            for (int j = 0; j < dimension; j++) {
                double sum = 0.0;
                if (ClusterSize > 0) {
                    for (int p = 0; p < ClusterSize; p++)
                        sum += clusters[i].getPoint(p).getVal(j);
                    clusters[i].setCentroidByPos(j, sum / ClusterSize);
                }
            }
        }

        if (done || iter >= iters) {
//            cout << "Clustering completed in iteration : " << iter << endl << endl;
            break;
        }
        iter++;
    }

//    //Print pointIds in each cluster
//    for (int i = 0; i < K; i++) {
//        cout << "Points in cluster " << clusters[i].getId() << " : ";
//        for (int j = 0; j < clusters[i].getSize(); j++) {
//            cout << clusters[i].getPoint(j).getId() << " ";
//        }
//        cout << endl;
//    }
//    cout << "========================" << endl;
}

int KMeans::readData(const string &fileName, vector<Point> *points) {
    ifstream infile;
    infile.open(fileName, ios::in);
    if (infile.is_open()) {
        int pointId = 1;
        string line;
        while (getline(infile, line)) {
            points->emplace_back(Point(pointId, line));
            pointId++;
        }
        infile.close();
        return SUCCESS_OPEN_FILE;
    } else {
        return UNABLE_OPEN_FILE;
    }
}

string KMeans::centroidsToString() {
    stringstream ss;
    for (int i = 0; i < K; ++i) {
        ss << clusters[i].clusterId << " ";
        for (double j : clusters[i].centroid) {
            ss << j << " ";
        }
        ss << endl;
    }
    ss << endl;
    return ss.str();
}

vector<int> KMeans::prediction(const vector<Point> &points) {
    vector<int> result;
    for (const auto &p : points) {
        int id = getNearestClusterId(p);
        result.push_back(id);
    }
    return result;
}
